package java.util;

import java.util.function.DoubleConsumer;
import java.util.function.LongConsumer;
import java.util.function.IntConsumer;
import java.util.stream.DoubleStream;
import java.util.stream.LongStream;
import java.util.stream.StreamSupport;
import java.util.stream.IntStream;
import java.security.SecureRandom;
import java.security.AccessController;
import sun.security.action.GetPropertyAction;
import java.util.concurrent.atomic.AtomicLong;

public final class SplittableRandom
{
    private static final long GOLDEN_GAMMA = -7046029254386353131L;
    private static final double DOUBLE_UNIT = 1.1102230246251565E-16;
    private long seed;
    private final long gamma;
    private static final AtomicLong defaultGen;
    static final String BadBound = "bound must be positive";
    static final String BadRange = "bound must be greater than origin";
    static final String BadSize = "size must be non-negative";
    
    private SplittableRandom(long seed, long gamma) {
        this.seed = seed;
        this.gamma = gamma;
    }
    
    private static long mix64(long n) {
        n = (n ^ n >>> 30) * -4658895280553007687L;
        n = (n ^ n >>> 27) * -7723592293110705685L;
        return n ^ n >>> 31;
    }
    
    private static int mix32(long n) {
        n = (n ^ n >>> 33) * 7109453100751455733L;
        return (int)((n ^ n >>> 28) * -3808689974395783757L >>> 32);
    }
    
    private static long mixGamma(long n) {
        n = (n ^ n >>> 33) * -49064778989728563L;
        n = (n ^ n >>> 33) * -4265267296055464877L;
        n = ((n ^ n >>> 33) | 0x1L);
        return (Long.bitCount(n ^ n >>> 1) < 24) ? (n ^ 0xAAAAAAAAAAAAAAAAL) : n;
    }
    
    private long nextSeed() {
        return this.seed += this.gamma;
    }
    
    private static long initialSeed() {
        String s = (String)AccessController.doPrivileged(new GetPropertyAction("java.util.secureRandomSeed"));
        if (s != null && s.equalsIgnoreCase("true")) {
            byte[] seed = SecureRandom.getSeed(8);
            long n = (long)seed[0] & 0xFFL;
            for (int i = 1; i < 8; ++i) {
                n = (n << 8 | ((long)seed[i] & 0xFFL));
            }
            return n;
        }
        return mix64(System.currentTimeMillis()) ^ mix64(System.nanoTime());
    }
    
    final long internalNextLong(long n, long n2) {
        long n3 = mix64(this.nextSeed());
        if (n < n2) {
            long n4 = n2 - n;
            long n5 = n4 - 1L;
            if ((n4 & n5) == 0x0L) {
                n3 = (n3 & n5) + n;
            }
            else if (n4 > 0L) {
                long n7;
                for (long n6 = n3 >>> 1; n6 + n5 - (n7 = n6 % n4) < 0L; n6 = mix64(this.nextSeed()) >>> 1) {}
                n3 = n7 + n;
            }
            else {
                while (n3 < n || n3 >= n2) {
                    n3 = mix64(this.nextSeed());
                }
            }
        }
        return n3;
    }
    
    final int internalNextInt(int n, int n2) {
        int n3 = mix32(this.nextSeed());
        if (n < n2) {
            int n4 = n2 - n;
            int n5 = n4 - 1;
            if ((n4 & n5) == 0x0) {
                n3 = (n3 & n5) + n;
            }
            else if (n4 > 0) {
                int n7;
                for (int n6 = n3 >>> 1; n6 + n5 - (n7 = n6 % n4) < 0; n6 = mix32(this.nextSeed()) >>> 1) {}
                n3 = n7 + n;
            }
            else {
                while (n3 < n || n3 >= n2) {
                    n3 = mix32(this.nextSeed());
                }
            }
        }
        return n3;
    }
    
    final double internalNextDouble(double n, double value) {
        double longBitsToDouble = (double)(this.nextLong() >>> 11) * 1.1102230246251565E-16;
        if (n < value) {
            longBitsToDouble = longBitsToDouble * (value - n) + n;
            if (longBitsToDouble >= value) {
                longBitsToDouble = Double.longBitsToDouble(Double.doubleToLongBits(value) - 1L);
            }
        }
        return longBitsToDouble;
    }
    
    public SplittableRandom(long n) {
        this(n, -7046029254386353131L);
    }
    
    public SplittableRandom() {
        long andAdd = SplittableRandom.defaultGen.getAndAdd(4354685564936845354L);
        this.seed = mix64(andAdd);
        this.gamma = mixGamma(andAdd - 7046029254386353131L);
    }
    
    public SplittableRandom split() {
        return new SplittableRandom(this.nextLong(), mixGamma(this.nextSeed()));
    }
    
    public int nextInt() {
        return mix32(this.nextSeed());
    }
    
    public int nextInt(int n) {
        if (n <= 0) {
            throw new IllegalArgumentException("bound must be positive");
        }
        int mix32 = mix32(this.nextSeed());
        int n2 = n - 1;
        int n4;
        if ((n & n2) == 0x0) {
            n4 = (mix32 & n2);
        }
        else {
            for (int n3 = mix32 >>> 1; n3 + n2 - (n4 = n3 % n) < 0; n3 = mix32(this.nextSeed()) >>> 1) {}
        }
        return n4;
    }
    
    public int nextInt(int n, int n2) {
        if (n >= n2) {
            throw new IllegalArgumentException("bound must be greater than origin");
        }
        return this.internalNextInt(n, n2);
    }
    
    public long nextLong() {
        return mix64(this.nextSeed());
    }
    
    public long nextLong(long n) {
        if (n <= 0L) {
            throw new IllegalArgumentException("bound must be positive");
        }
        long mix64 = mix64(this.nextSeed());
        long n2 = n - 1L;
        long n4;
        if ((n & n2) == 0x0L) {
            n4 = (mix64 & n2);
        }
        else {
            for (long n3 = mix64 >>> 1; n3 + n2 - (n4 = n3 % n) < 0L; n3 = mix64(this.nextSeed()) >>> 1) {}
        }
        return n4;
    }
    
    public long nextLong(long n, long n2) {
        if (n >= n2) {
            throw new IllegalArgumentException("bound must be greater than origin");
        }
        return this.internalNextLong(n, n2);
    }
    
    public double nextDouble() {
        return (double)(mix64(this.nextSeed()) >>> 11) * 1.1102230246251565E-16;
    }
    
    public double nextDouble(double value) {
        if (value <= 0.0) {
            throw new IllegalArgumentException("bound must be positive");
        }
        double n = (double)(mix64(this.nextSeed()) >>> 11) * 1.1102230246251565E-16 * value;
        return (n < value) ? n : Double.longBitsToDouble(Double.doubleToLongBits(value) - 1L);
    }
    
    public double nextDouble(double n, double n2) {
        if (n >= n2) {
            throw new IllegalArgumentException("bound must be greater than origin");
        }
        return this.internalNextDouble(n, n2);
    }
    
    public boolean nextBoolean() {
        return mix32(this.nextSeed()) < 0;
    }
    
    public IntStream ints(long n) {
        if (n < 0L) {
            throw new IllegalArgumentException("size must be non-negative");
        }
        return StreamSupport.intStream(new SplittableRandom.RandomIntsSpliterator(this, 0L, n, Integer.MAX_VALUE, 0), false);
    }
    
    public IntStream ints() {
        return StreamSupport.intStream(new SplittableRandom.RandomIntsSpliterator(this, 0L, Long.MAX_VALUE, Integer.MAX_VALUE, 0), false);
    }
    
    public IntStream ints(long n, int n2, int n3) {
        if (n < 0L) {
            throw new IllegalArgumentException("size must be non-negative");
        }
        if (n2 >= n3) {
            throw new IllegalArgumentException("bound must be greater than origin");
        }
        return StreamSupport.intStream(new SplittableRandom.RandomIntsSpliterator(this, 0L, n, n2, n3), false);
    }
    
    public IntStream ints(int n, int n2) {
        if (n >= n2) {
            throw new IllegalArgumentException("bound must be greater than origin");
        }
        return StreamSupport.intStream(new SplittableRandom.RandomIntsSpliterator(this, 0L, Long.MAX_VALUE, n, n2), false);
    }
    
    public LongStream longs(long n) {
        if (n < 0L) {
            throw new IllegalArgumentException("size must be non-negative");
        }
        return StreamSupport.longStream(new SplittableRandom.RandomLongsSpliterator(this, 0L, n, Long.MAX_VALUE, 0L), false);
    }
    
    public LongStream longs() {
        return StreamSupport.longStream(new SplittableRandom.RandomLongsSpliterator(this, 0L, Long.MAX_VALUE, Long.MAX_VALUE, 0L), false);
    }
    
    public LongStream longs(long n, long n2, long n3) {
        if (n < 0L) {
            throw new IllegalArgumentException("size must be non-negative");
        }
        if (n2 >= n3) {
            throw new IllegalArgumentException("bound must be greater than origin");
        }
        return StreamSupport.longStream(new SplittableRandom.RandomLongsSpliterator(this, 0L, n, n2, n3), false);
    }
    
    public LongStream longs(long n, long n2) {
        if (n >= n2) {
            throw new IllegalArgumentException("bound must be greater than origin");
        }
        return StreamSupport.longStream(new SplittableRandom.RandomLongsSpliterator(this, 0L, Long.MAX_VALUE, n, n2), false);
    }
    
    public DoubleStream doubles(long n) {
        if (n < 0L) {
            throw new IllegalArgumentException("size must be non-negative");
        }
        return StreamSupport.doubleStream(new SplittableRandom.RandomDoublesSpliterator(this, 0L, n, Double.MAX_VALUE, 0.0), false);
    }
    
    public DoubleStream doubles() {
        return StreamSupport.doubleStream(new SplittableRandom.RandomDoublesSpliterator(this, 0L, Long.MAX_VALUE, Double.MAX_VALUE, 0.0), false);
    }
    
    public DoubleStream doubles(long n, double n2, double n3) {
        if (n < 0L) {
            throw new IllegalArgumentException("size must be non-negative");
        }
        if (n2 >= n3) {
            throw new IllegalArgumentException("bound must be greater than origin");
        }
        return StreamSupport.doubleStream(new SplittableRandom.RandomDoublesSpliterator(this, 0L, n, n2, n3), false);
    }
    
    public DoubleStream doubles(double n, double n2) {
        if (n >= n2) {
            throw new IllegalArgumentException("bound must be greater than origin");
        }
        return StreamSupport.doubleStream(new SplittableRandom.RandomDoublesSpliterator(this, 0L, Long.MAX_VALUE, n, n2), false);
    }
    
    static {
        defaultGen = new AtomicLong(initialSeed());
    }
    
    static final class RandomIntsSpliterator implements Spliterator.OfInt
    {
        final SplittableRandom rng;
        long index;
        final long fence;
        final int origin;
        final int bound;
        
        RandomIntsSpliterator(SplittableRandom rng, long index, long fence, int origin, int bound) {
            this.rng = rng;
            this.index = index;
            this.fence = fence;
            this.origin = origin;
            this.bound = bound;
        }
        
        @Override
        public SplittableRandom.RandomIntsSpliterator trySplit() {
            long index = this.index;
            long index2 = index + this.fence >>> 1;
            return (index2 <= index) ? null : new SplittableRandom.RandomIntsSpliterator(this.rng.split(), index, this.index = index2, this.origin, this.bound);
        }
        
        @Override
        public long estimateSize() {
            return this.fence - this.index;
        }
        
        @Override
        public int characteristics() {
            return 17728;
        }
        
        @Override
        public boolean tryAdvance(IntConsumer intConsumer) {
            if (intConsumer == null) {
                throw new NullPointerException();
            }
            long index = this.index;
            if (index < this.fence) {
                intConsumer.accept(this.rng.internalNextInt(this.origin, this.bound));
                this.index = index + 1L;
                return true;
            }
            return false;
        }
        
        @Override
        public void forEachRemaining(IntConsumer intConsumer) {
            if (intConsumer == null) {
                throw new NullPointerException();
            }
            long index = this.index;
            long fence = this.fence;
            if (index < fence) {
                this.index = fence;
                SplittableRandom rng = this.rng;
                int origin = this.origin;
                int bound = this.bound;
                do {
                    intConsumer.accept(rng.internalNextInt(origin, bound));
                } while (++index < fence);
            }
            return;
        }
    }
    
    static final class RandomLongsSpliterator implements Spliterator.OfLong
    {
        final SplittableRandom rng;
        long index;
        final long fence;
        final long origin;
        final long bound;
        
        RandomLongsSpliterator(SplittableRandom rng, long index, long fence, long origin, long bound) {
            this.rng = rng;
            this.index = index;
            this.fence = fence;
            this.origin = origin;
            this.bound = bound;
        }
        
        @Override
        public SplittableRandom.RandomLongsSpliterator trySplit() {
            long index = this.index;
            long index2 = index + this.fence >>> 1;
            return (index2 <= index) ? null : new SplittableRandom.RandomLongsSpliterator(this.rng.split(), index, this.index = index2, this.origin, this.bound);
        }
        
        @Override
        public long estimateSize() {
            return this.fence - this.index;
        }
        
        @Override
        public int characteristics() {
            return 17728;
        }
        
        @Override
        public boolean tryAdvance(LongConsumer longConsumer) {
            if (longConsumer == null) {
                throw new NullPointerException();
            }
            long index = this.index;
            if (index < this.fence) {
                longConsumer.accept(this.rng.internalNextLong(this.origin, this.bound));
                this.index = index + 1L;
                return true;
            }
            return false;
        }
        
        @Override
        public void forEachRemaining(LongConsumer longConsumer) {
            if (longConsumer == null) {
                throw new NullPointerException();
            }
            long index = this.index;
            long fence = this.fence;
            if (index < fence) {
                this.index = fence;
                SplittableRandom rng = this.rng;
                long origin = this.origin;
                long bound = this.bound;
                do {
                    longConsumer.accept(rng.internalNextLong(origin, bound));
                } while (++index < fence);
            }
            return;
        }
    }
    
    static final class RandomDoublesSpliterator implements Spliterator.OfDouble
    {
        final SplittableRandom rng;
        long index;
        final long fence;
        final double origin;
        final double bound;
        
        RandomDoublesSpliterator(SplittableRandom rng, long index, long fence, double origin, double bound) {
            this.rng = rng;
            this.index = index;
            this.fence = fence;
            this.origin = origin;
            this.bound = bound;
        }
        
        @Override
        public SplittableRandom.RandomDoublesSpliterator trySplit() {
            long index = this.index;
            long index2 = index + this.fence >>> 1;
            return (index2 <= index) ? null : new SplittableRandom.RandomDoublesSpliterator(this.rng.split(), index, this.index = index2, this.origin, this.bound);
        }
        
        @Override
        public long estimateSize() {
            return this.fence - this.index;
        }
        
        @Override
        public int characteristics() {
            return 17728;
        }
        
        @Override
        public boolean tryAdvance(DoubleConsumer doubleConsumer) {
            if (doubleConsumer == null) {
                throw new NullPointerException();
            }
            long index = this.index;
            if (index < this.fence) {
                doubleConsumer.accept(this.rng.internalNextDouble(this.origin, this.bound));
                this.index = index + 1L;
                return true;
            }
            return false;
        }
        
        @Override
        public void forEachRemaining(DoubleConsumer doubleConsumer) {
            if (doubleConsumer == null) {
                throw new NullPointerException();
            }
            long index = this.index;
            long fence = this.fence;
            if (index < fence) {
                this.index = fence;
                SplittableRandom rng = this.rng;
                double origin = this.origin;
                double bound = this.bound;
                do {
                    doubleConsumer.accept(rng.internalNextDouble(origin, bound));
                } while (++index < fence);
            }
            return;
        }
    }
}
